@@ -12,9 +12,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library; if not, see <http://www.gnu.org/licenses/>.
  *
  * Author: Ryan Lortie <desrt@desrt.ca>
  */
@@ -56,6 +54,7 @@
 struct _DConfChangeset
 {
   GHashTable *table;
+  GHashTable *dir_resets;
   guint is_database : 1;
   guint is_sealed : 1;
   gint ref_count;
@@ -77,7 +76,7 @@ unref_gvariant0 (gpointer data)
  *
  * Creates a new, empty, #DConfChangeset.
  *
- * Returns: the new #DConfChangeset.
+ * Returns: (transfer full): the new #DConfChangeset.
  **/
 DConfChangeset *
 dconf_changeset_new (void)
@@ -93,7 +92,7 @@ dconf_changeset_new (void)
 
 /**
  * dconf_changeset_new_database:
- * @copy_of: (allow-none): a #DConfChangeset to copy
+ * @copy_of: (nullable): a #DConfChangeset to copy
  *
  * Creates a new #DConfChangeset in "database" mode, possibly
  * initialising it with the values of another changeset.
@@ -115,7 +114,7 @@ dconf_changeset_new (void)
  * If @copy_of is non-%NULL then its contents will be copied into the
  * created changeset.  @copy_of must be a database-mode changeset.
  *
- * Returns: a new #DConfChangeset in "database" mode
+ * Returns: (transfer full): a new #DConfChangeset in "database" mode
  *
  * Since: 0.16
  */
@@ -159,6 +158,9 @@ dconf_changeset_unref (DConfChangeset *changeset)
 
       g_hash_table_unref (changeset->table);
 
+      if (changeset->dir_resets)
+        g_hash_table_unref (changeset->dir_resets);
+
       g_slice_free (DConfChangeset, changeset);
     }
 }
@@ -169,7 +171,7 @@ dconf_changeset_unref (DConfChangeset *changeset)
  *
  * Increases the reference count on @changeset
  *
- * Returns: @changeset
+ * Returns: (transfer full): @changeset
  **/
 DConfChangeset *
 dconf_changeset_ref (DConfChangeset *changeset)
@@ -179,11 +181,27 @@ dconf_changeset_ref (DConfChangeset *changeset)
   return changeset;
 }
 
+static void
+dconf_changeset_record_dir_reset (DConfChangeset *changeset,
+                                  const gchar    *dir)
+{
+  g_return_if_fail (dconf_is_dir (dir, NULL));
+  g_return_if_fail (!changeset->is_database);
+  g_return_if_fail (!changeset->is_sealed);
+
+  if (!changeset->dir_resets)
+    changeset->dir_resets = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+
+  g_hash_table_insert (changeset->table, g_strdup (dir), NULL);
+  g_hash_table_add (changeset->dir_resets, g_strdup (dir));
+}
+
 /**
  * dconf_changeset_set:
  * @changeset: a #DConfChangeset
  * @path: a path to modify
- * @value: the value for the key, or %NULL to reset
+ * @value: (nullable): the value for the key, or %NULL to reset. If it has a
+ *  floating reference it's consumed.
  *
  * Adds an operation to modify @path to a #DConfChangeset.
  *
@@ -219,7 +237,7 @@ dconf_changeset_set (DConfChangeset *changeset,
 
       /* If this is a non-database then record the reset itself. */
       if (!changeset->is_database)
-        g_hash_table_insert (changeset->table, g_strdup (path), NULL);
+        dconf_changeset_record_dir_reset (changeset, path);
     }
 
   /* ...or a value reset */
@@ -243,7 +261,7 @@ dconf_changeset_set (DConfChangeset *changeset,
  * dconf_changeset_get:
  * @changeset: a #DConfChangeset
  * @key: the key to check
- * @value: a return location for the value, or %NULL
+ * @value: (transfer full) (optional) (nullable): a return location for the value, or %NULL
  *
  * Checks if a #DConfChangeset has an outstanding request to change
  * the value of the given @key.
@@ -265,7 +283,26 @@ dconf_changeset_get (DConfChangeset  *changeset,
   gpointer tmp;
 
   if (!g_hash_table_lookup_extended (changeset->table, key, NULL, &tmp))
-    return FALSE;
+    {
+      /* Did not find an exact match, so check for dir resets */
+      if (changeset->dir_resets)
+        {
+          GHashTableIter iter;
+          gpointer dir;
+
+          g_hash_table_iter_init (&iter, changeset->dir_resets);
+          while (g_hash_table_iter_next (&iter, &dir, NULL))
+            if (g_str_has_prefix (key, dir))
+              {
+                if (value)
+                  *value = NULL;
+
+                return TRUE;
+              }
+        }
+
+      return FALSE;
+    }
 
   if (value)
     *value = tmp ? g_variant_ref (tmp) : NULL;
@@ -315,7 +352,7 @@ dconf_changeset_is_similar_to (DConfChangeset *changeset,
 /**
  * DConfChangesetPredicate:
  * @path: a path, as per dconf_is_path()
- * @value: a #GVariant, or %NULL
+ * @value: (nullable): a #GVariant, or %NULL
  * @user_data: user data pointer
  *
  * Callback function type for predicates over items in a
@@ -521,9 +558,9 @@ dconf_changeset_seal (DConfChangeset *changeset)
 /**
  * dconf_changeset_describe:
  * @changeset: a #DConfChangeset
- * @prefix: the prefix under which changes have been requested
- * @paths: the list of paths changed, relative to @prefix
- * @values: the list of values changed
+ * @prefix: (transfer none) (optional) (out): the prefix under which changes have been requested
+ * @paths: (transfer none) (optional) (out): the list of paths changed, relative to @prefix
+ * @values: (transfer none) (optional) (out): the list of values changed
  *
  * Describes @changeset.
  *
@@ -574,7 +611,7 @@ dconf_changeset_describe (DConfChangeset       *changeset,
  * The returned value has no particular format and should only be passed
  * to dconf_changeset_deserialise().
  *
- * Returns: a floating #GVariant
+ * Returns: (transfer full): a floating #GVariant
  **/
 GVariant *
 dconf_changeset_serialise (DConfChangeset *changeset)
@@ -594,7 +631,7 @@ dconf_changeset_serialise (DConfChangeset *changeset)
 
 /**
  * dconf_changeset_deserialise:
- * @serialised: a #GVariant from dconf_changeset_serialise()
+ * @serialised: (transfer none): a #GVariant from dconf_changeset_serialise()
  *
  * Creates a #DConfChangeset according to a serialised description
  * returned from an earlier call to dconf_changeset_serialise().
@@ -605,7 +642,7 @@ dconf_changeset_serialise (DConfChangeset *changeset)
  * This call never fails, even if @serialised is not in the correct
  * format.  Improperly-formatted parts are simply ignored.
  *
- * Returns: a new #DConfChangeset
+ * Returns: (transfer full): a new #DConfChangeset
  **/
 DConfChangeset *
 dconf_changeset_deserialise (GVariant *serialised)
@@ -626,16 +663,11 @@ dconf_changeset_deserialise (GVariant *serialised)
        *
        * If we get an invalid case, just fall through and ignore it.
        */
-      if (value == NULL)
-        {
-          if (dconf_is_path (key, NULL))
-            g_hash_table_insert (changeset->table, g_strdup (key), NULL);
-        }
-      else
-        {
-          if (dconf_is_key (key, NULL))
-            g_hash_table_insert (changeset->table, g_strdup (key), g_variant_ref (value));
-        }
+      if (dconf_is_key (key, NULL))
+        g_hash_table_insert (changeset->table, g_strdup (key), value ? g_variant_ref (value) : NULL);
+
+      else if (dconf_is_dir (key, NULL) && value == NULL)
+        dconf_changeset_record_dir_reset (changeset, key);
     }
 
   return changeset;
@@ -644,7 +676,8 @@ dconf_changeset_deserialise (GVariant *serialised)
 /**
  * dconf_changeset_new_write:
  * @path: a dconf path
- * @value: a #GVariant, or %NULL
+ * @value: (nullable): a #GVariant, or %NULL. If it has a floating reference it's
+ *  consumed.
  *
  * Creates a new #DConfChangeset with one change.  This is equivalent to
  * calling dconf_changeset_new() and then dconf_changeset_set() with
@@ -752,7 +785,7 @@ dconf_changeset_change (DConfChangeset *changeset,
  * dconf_changeset_change() will result in the two changesets being
  * equal.
  *
- * Returns: (transfer full): the changes, or %NULL
+ * Returns: (transfer full) (nullable): the changes, or %NULL
  *
  * Since: 0.16
  */
